#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include "eval.hpp"

using namespace aed;
using namespace std;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// EJERCICIO 1. INT_SET_T
class odd_t : public int_set_t {
public:
  bool isin(int x) { return modulo(x,2)==1; }
};
#define HAS_ODD_T

class even_t : public int_set_t {
public:
  bool isin(int x) {
    // cout << "even_t, x " << x << endl;
    return modulo(x,2)==0;
  }
};
#define HAS_EVEN_T

class all_t : public int_set_t {
public:
  bool isin(int x) { return true; }
};
#define HAS_ALL_T

class empty_t : public int_set_t {
public:
  bool isin(int x) { return false; }
};
#define HAS_EMPTY_T

class positive_t : public int_set_t {
public:
  bool isin(int x) { return x>0; }
};
#define HAS_POSITIVE_T

class stride_t : public int_set_t {
public:
  int start,end,step;
  bool isin(int x) { return x>=start && x<end && modulo(x-start,step)==0; }
};
#define HAS_STRIDE_T

class vector_t : public int_set_t {
public:
  vector<int> w;
  bool isin(int x) {
    for (auto q : w) if (q==x) return true;
    return false;
  }
};
#define HAS_VECTOR_T
class intersection_t : public int_set_t {
public:
  unique_ptr<int_set_t> s1,s2;
  bool isin(int x) { return s1->isin(x) && s2->isin(x); }
};
#define HAS_INTERSECTION_T

class union_t : public int_set_t {
public:
  unique_ptr<int_set_t> s1,s2;
  bool isin(int x) { return s1->isin(x) || s2->isin(x); }
};
#define HAS_UNION_T

class difference_t : public int_set_t {
public:
  unique_ptr<int_set_t> s1,s2;
  bool isin(int x) { return s1->isin(x) && ! s2->isin(x); }
};
#define HAS_DIFFERENCE_T

class complement_t : public int_set_t {
public:
  unique_ptr<int_set_t> s1;
  bool isin(int x) { return !s1->isin(x); }
};
#define HAS_COMPLEMENT_T

// Otra implementacion de EVEN_T como complemento de ODD_T
class even2_t : public complement_t {
public:
  even2_t() { s1.reset(new odd_t); }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
// EJERCICIO 2. PERM_T/NILPOT
class shift1_t : public perm_t {
public:
  void mix(vector<int> &w) {
    int N=w.size();
    vector<int> out = w;
    for (int j=0; j<N-1; j++) {
      out[j+1] = w[j];
    }
    out[0] = w[N-1];
    w = out;
  }
};
#define HAS_SHIFT1_T

class shiftm_t : public perm_t {
public:
  int m;
  void mix(vector<int> &w) {
    int n=w.size();
    vector<int> out = w;
    for (int j=0; j<n; j++) {
      out[modulo(j+m,n)] = w[j];
    }
    w = out;
  }
};
#define HAS_SHIFTM_T

class oddeven_t : public perm_t {
public:
  int m;
  void mix(vector<int> &w) {
    int n = w.size();
    vector<int> out = w;
    for (int j=0; j+1<n; j+=2) {
      out[j+1] = w[j];
      out[j] = w[j+1];
    }
    w = out;
  }
};
#define HAS_ODDEVEN_T

class reflex_t : public perm_t {
public:
  void mix(vector<int> &w) {
    int n = w.size();
    int m = 2*(n/2);
    vector<int> out=w;
    for (int j=0; j<m; j++) {
      int j1=m-j-1;
      out[j1] = w[j];
    }
    w = out;
  }
};
#define HAS_REFLEX_T

class shiftm1m2_t : public perm_t {
public:
  int m1,m2;
  void mix(vector<int> &w) {
    int n = w.size();
    assert(m1+m2<=n);
    vector<int> out=w;
    int j1,j2;
    for (int j=0; j<m1; j++) {
      j1=j; j2=modulo(j+1,m1);
      out[j2]=w[j1];
    }
    for (int j=0; j<m2; j++) {
      j1=m1+j; j2=m1+modulo(j+1,m2);
      out[j2]=w[j1];
    }
    w = out;
  }
};
#define HAS_SHIFTM1M2_T

int nilpot(perm_t &perm,int N) {
  vector<int> w(N),w0(N);
  for (int j=0; j<N; j++) w0[j] = j;
  int n=0; w=w0;
  while (1) {
    perm.mix(w); n++;
    if (w==w0) return n;
  }
}

#include "bintset.hpp"

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  int vrbs = 0;
#if 1
  eval1_t ev1;
  ev1.evfun(vrbs);
#endif

#if 0
  eval2_t ev2;
  ev2.evfun(vrbs);
#endif
  return 0;
}
