#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include "eval.hpp"

using namespace aed;
using namespace std;
// COMPLETAR DNI y NOMBRE AQUI:
// Nombre: Esponja, Bob
int DNI=23456789;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
typedef vector<vector<int>> history_t;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int sumcards(vector<int> &cards) {
  int sum=0;
  for (auto &c : cards) sum += c;
  return sum;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class player_t {
public:
  virtual bool wants_card(vector<int> &hand,
                          history_t &me,history_t &you)=0;
  virtual void reset() { }
  virtual string description() { return "unknown"; }
};

int summax=21,maxcard=10;

int get_card() { return rand()%maxcard+1; }

bool mkhand(vector<int> &cards,player_t &p,
            history_t &me,history_t &you,int vrbs=0,
            const char *label=NULL) {
  cards.push_back(get_card());
  while (p.wants_card(cards,me,you)) {
    cards.push_back(get_card());
    if (sumcards(cards)>summax) break;
  }
  bool ok = sumcards(cards)<=summax;
  if (vrbs) {
    if (label) printf("%s: ",label);
    for (auto &c : cards) printf("%d ",c);
    printf(" ok %d\n",ok);
  }
  return ok;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
enum rslt_t {WINS1=0,DRAW=1,WINS2=2};
void playbjgame(player_t &p1,player_t &p2,
                vector<double> &stats,
                int nhands=1000) {
  stats.clear();
  stats.resize(3,0.0);
  p1.reset();
  p2.reset();
  history_t h1,h2;
  int vrbs=0;
  int busts1=0,nok1=0,oksum1=0;
  for (int j=0; j<nhands; j++) {
    vector<int> cards1,cards2;
    bool ok1=mkhand(cards1,p1,h1,h2,vrbs,"p1");
    bool ok2=mkhand(cards2,p2,h2,h1,vrbs,"p2");
    h1.push_back(cards1);
    h2.push_back(cards2);
    int sum1=sumcards(cards1);
    int sum2=sumcards(cards2);
    rslt_t r;
    if (!ok1 && !ok2) r=DRAW;
    else if (!ok1) r=WINS2;
    else if (!ok2) r=WINS1;
    else r=(sum1>sum2? WINS1 : sum2>sum1? WINS2 : DRAW);
    stats[r]+=1.0;
    busts1 += !ok1;
    nok1 += ok1;
    if (ok1) oksum1 += sum1;
  }
  // printf("%s bust %.3f cardsum %.3f\n",
  //        p1.description().c_str(),
  //        double(busts1)/nhands,
  //        double(oksum1)/nok1);
  for (int l=0; l<3; l++) stats[l] /= nhands;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class fixed_t : public player_t {
public:
  int ncards;
  fixed_t(int nc=4) : ncards(nc) {}
  bool wants_card(vector<int> &cards,
                  history_t &me,history_t &you) {
    return int(cards.size())<ncards;
  }
  string description() {
    return "fixed("+to_string(ncards)+")";
  }
};
#define HAS_FIXED_T

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class target_t : public player_t {
public:
  int target;
  target_t(int trg=18) : target(trg) {}
  bool wants_card(vector<int> &cards,
                  history_t &me,history_t &you) {
    return sumcards(cards)<target;
  }
  string description() {
    return "target("+to_string(target)+")";
  }
};
#define HAS_TARGET_T

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class adaptive_t : public player_t {
public:
  int target;
  adaptive_t() : target(18) {}
  bool wants_card(vector<int> &cards,
                  history_t &me,history_t &you) {
    // If size of history is large enough compute my mean
    // and your mean
    // Number of hands
    int nhands=you.size();
    if (nhands>100) {
      int youokmean=0,okhands=0;
      for (auto &c : you) {
        int sum = sumcards(c);
        if (sum<=summax) {
          youokmean += sum;
          okhands++;
        }
      }
      target = double(youokmean)/okhands;
      target = (target>18? 18 : target);
    }
    return sumcards(cards)<target;
  }
  string description() { return "adaptive"; }
};
#define HAS_ADAPTIVE_T

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class adaptive2_t : public player_t {
public:
  int target,oksum,okhands;
  adaptive2_t() : target(18), oksum(0), okhands(0) {}
  void reset() { oksum=0; okhands=0; }
  bool wants_card(vector<int> &cards,
                  history_t &me,history_t &you) {
    // If size of history is large enough compute my mean
    // and your mean
    // Number of hands
    int nhands=you.size();
    if (nhands>0) {
      auto &c = you.back();
      int sum = sumcards(c);
      if (sum<=summax) {
        oksum += sum;
        okhands++;
      }
    }
    if (nhands>100) {
      target = double(oksum)/okhands;
      target = (target>18 || target<13? 18 : target);
    }
    return sumcards(cards)<target;
  }
  string description() { return "adaptive2"; }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
typedef vector<double> vd_t;
typedef vector<vd_t> vvd_t;

class linearop_t {
public:
  // Computa y=A*x
  virtual void apply(vd_t &y,vd_t &x)=0;
  string description() { return "unknown"; }
};

typedef unique_ptr<linearop_t> plinearop_t;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
double dot(vd_t &x,vd_t &y) {
    int n=x.size();
    double sum=0.0;
    for (int j=0; j<n; j++) sum += y[j]*x[j];
    return sum;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
double norm(vd_t &x) { return sqrt(dot(x,x)); }

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void dumpv(vd_t &x,const char *label=NULL) {
  if (label) printf("%s: ",label);
  for (auto &w : x) printf("%f ",w);
  printf("\n");
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool is_spd(linearop_t &A,vvd_t &vectors,double tol=0.0) {
  int nv=vectors.size();
  vd_t y;
  for (int j=0; j<nv; j++) {
    auto &x = vectors[j];
    A.apply(y,x);
    // dumpv(x,"x");
    // dumpv(y,"y");
    // printf("dot(x,y) %g\n",dot(x,y));
    if (dot(x,y)<-tol) return false;
  }
  return true;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool is_normpres(linearop_t &A,vvd_t &vectors,double tol=0.0) {
  int nv=vectors.size();
  vd_t y;
  for (int j=0; j<nv; j++) {
    auto &x = vectors[j];
    A.apply(y,x);
    double ax=norm(x),ay=norm(y);
    if (fabs(ax-ay)>tol) return false;
  }
  return true;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool are_equal(linearop_t &A1,linearop_t &A2,vvd_t &vectors,double tol=0.0) {
  int nv=vectors.size();
  vd_t y1,y2;
  for (int j=0; j<nv; j++) {
    auto &x = vectors[j];
    int m=x.size();
    A1.apply(y1,x);
    A2.apply(y2,x);
    double d=0.0;
    for (int j=0; j<m; j++) d += fabs(y1[j]-y2[j]);
    if (d>tol) return false;
  }
  return true;
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class identity_t : public linearop_t {
public:
  void apply(vd_t &y,vd_t &x) { y=x; }
  string description() { return "Id"; }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class scale_t : public linearop_t {
  identity_t id;
  double a;
public:
  scale_t(double a_=1.0) : a(a_) {}
  void apply(vd_t &y,vd_t &x) {
    y = x;
    int n=x.size();
    for (int j=0; j<n; j++) y[j] *= a;
  }
  string description() { return "scale("+to_string(a)+")"; }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class rot_t : public linearop_t {
  int shift;
public:
  rot_t(int s=1) : shift(s) {}
  void apply(vd_t &y,vd_t &x) {
    int n=x.size();
    assert(n>=2);
    y=x;
    for (int j=0; j<n; j++)
      y[(j+shift)%n] = x[j];
  }
  string description() {
    return "rot("+to_string(shift)+")";
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class reflex_t : public linearop_t {
  int axis;
public:
  reflex_t(int ax=0) : axis(ax) {}
  void apply(vd_t &y,vd_t &x) {
    int n=x.size();
    assert(axis<n);
    y=x;
    y[axis] = -x[axis];
  }
  string description() {
    return "axis("+to_string(axis)+")";
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class compose_t : public linearop_t {
public:
  vector<plinearop_t> ops;
  void apply(vd_t &y,vd_t &x) {
    y = x;
    auto z = x;
    for (auto &p : ops) {
      p->apply(y,z);
      z = y;
    }
  }
  string description() {
    string d="(";
    for (auto &p : ops) d += p->description()+",";
    d += ")";
    return d;
  }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mkrandvectors(vvd_t &vectors,int dim=3,int nv=100) {
  for (int j=0; j<nv; j++) {
    vd_t x(dim);
    for (int k=0; k<dim; k++) x[k]=2*drand()-1.0;
    vectors.push_back(x);
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main_linops() {
  vvd_t vectors;
  mkrandvectors(vectors);

  identity_t id;
  scale_t sc5(5),scm5(-5),sc02(0.2);
  rot_t rot;
  reflex_t reflex(1);

  printf("spd(id) %d\n",is_spd(id,vectors));
  printf("spd(scale(5)) %d\n",is_spd(sc5,vectors));
  printf("spd(scale(-5)) %d\n",is_spd(scm5,vectors));
  printf("spd(rot) %d\n",is_spd(rot,vectors));
  printf("spd(reflex(1)) %d\n",is_spd(reflex,vectors));
  
#if 0
  printf("normpres(id) %d\n",is_normpres(id,vectors,1e-10));
  printf("normpres(scale(5)) %d\n",is_normpres(sc5,vectors,1e-10));
  printf("normpres(scale(-5)) %d\n",is_normpres(scm5,vectors,1e-10));
  printf("normpres(reflex(1)) %d\n",is_normpres(reflex,vectors));

  compose_t id2;
  id2.ops.emplace_back(new scale_t(5));
  id2.ops.emplace_back(new scale_t(0.2));
  printf("normpres(id2) %d\n",is_normpres(id2,vectors,1e-10));

  compose_t id3;
  id2.ops.emplace_back(new rot_t(1));
  id2.ops.emplace_back(new rot_t(1));
  id2.ops.emplace_back(new rot_t(1));
  printf("normpres(id3) %d\n",is_normpres(id3,vectors,1e-10));

  printf("are_equal(id1,id2) %d\n",are_equal(id,id2,vectors,1e-10));
  printf("are_equal(id1,id3) %d\n",are_equal(id,id2,vectors,1e-10));

  scale_t sc25(25.0);
  compose_t sc5x5;
  sc5x5.ops.emplace_back(new scale_t(5.0));
  sc5x5.ops.emplace_back(new scale_t(5.0));
  printf("are_equal(sc5x5,sc25) %d\n",are_equal(sc5x5,sc25,vectors,1e-10));
#endif  
  return 0;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  fixed_t p1(3);
  for (int k=2; k<=6; k++) {
    fixed_t p2(k);
    vector<double> stats;
    playbjgame(p1,p2,stats,100000);
    printf("fix(%d) %f, draw %f, fix(%d) %f\n",
           3,stats[0],stats[1],k,stats[2]);
  }    
#if 0  
  vvd_t vectors;
  mkrandvectors(vectors);
  scale_t smas(1.0),smenos(-1.0);
  cout << "is_spd(smas) " << is_spd(smas,vectors) << endl;
  cout << "is_spd(smenos) " << is_spd(smenos,vectors) << endl;
  return 0;
#endif
  // return main_linops();
}
