#include "eval.hpp"
#include <cstdlib>

using namespace aed;
using namespace std;

int kill(int j) { return (j+1)%3; }

class ppt_player_t {
public:
  virtual int play(vector<int> &you,vector<int> &me)=0;
};

vector<int>
ppt_game(ppt_player_t &p1,ppt_player_t &p2,int Nrounds=1000) {
  int j1,j2;
  int win1=0, win2=0, draw=0;
  vector<int> plays1,plays2;
  for (int j=0; j<Nrounds; j++) {
    j1 = p1.play(plays2,plays1);
    j2 = p2.play(plays1,plays2);
    // printf("%d %d\n",j1,j2);
    if (j1==j2) draw++;
    else if ((j1+1)%3==j2) win2++;
    else win1++;
    plays1.push_back(j1);
    plays2.push_back(j2);
  }
  vector<int> rslt;
  rslt.push_back(win1);
  rslt.push_back(win2);
  rslt.push_back(draw);
  return rslt;
};

class random_t : public ppt_player_t {
public:
  virtual int play(vector<int> &you,
                        vector<int> &me) {
    return rand()%3;
  }
};

class fixed_t : public ppt_player_t {
public:
  virtual int play(vector<int> &you,
                        vector<int> &me) {
    return 1;     
  }
};

class repeater_t : public ppt_player_t {
public:
  virtual int play(vector<int> &you,
                        vector<int> &me) {
    int N = me.size();
    if (N==0) return rand()%3;
    else {
      int last = me[N-1];
      return (last+1+rand()%2)%3;
    }
  }
};

class repeater_killer_t : public ppt_player_t {
public:
  virtual int play(vector<int> &you,
                        vector<int> &me) {
    int N = you.size();
    if (N==0) return rand()%3;
    else {
      int last = you[N-1];
      // return (last+rand()%2)%3;
      return (last+2)%3;
    }
  }
};

bool ispermfun(set<int> &S,map_fun_t f) {
  set<int> fS;
  set<int>::iterator q = S.begin();
  int N = S.size();
  while (q!=S.end()) fS.insert(f(*q++));
  return fS==S;
}

void map2freq(list< set<int> > &LS, map<int,int> &freq) {
  freq.clear();
  typedef list< set<int> > ls_t;
  ls_t::iterator q = LS.begin();
  while (q!=LS.end()) {
    set<int> &S = *q++;
    set<int>::iterator s = S.begin();
    while (s!= S.end()) freq[*s++]++;
  }
}

void filtbyimg(map<int,int> &M,map<int,int> &M2,pred_t pred) {
  M2.clear();
  map<int,int>::iterator q = M.begin();
  while (q!=M.end()) {
    if (pred(q->second)) M2[q->first] = q->second;
    q++;
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  Eval ev;
  int vrbs = 1;

#if 1
  random_t p1,p2;
  fixed_t f1,f2;
  repeater_t r1,r2;
  repeater_killer_t rk1,rk2;
  vector<int> rslt;

  
#define GAME(p1,p2,pt1,pt2)                     \
  rslt = ppt_game(p1,p2);                       \
  printf("%s %d, %s %d, draw %d\n",            \
         pt1,rslt[0],pt2,rslt[1],rslt[2]);
                          
  GAME(p1,p2,"random","random");
  GAME(p1,f2,"random","fixed");
  GAME(f1,f2,"fixed","fixed");

  GAME(r1,r2,"repeater","repeater");
  GAME(r1,rk2,"repeater","repeater_killer");
  GAME(rk1,rk2,"repeater_killer","repeater_killer");
  GAME(f1,rk1,"fixed","repeater_killer");
#endif

#if 0
  set<int> S1;
  for (int j=0; j<N1; j++) S1.insert(j);
  cout << "isperm f1,S1: " << ispermfun(S1,f1) << endl;
#endif

#if 0  
  list< set<int> > LS;
  set<int> S;
  for (int j=0; j<5; j++) {
    S.clear();
    for (int k=j; k<5; k++) S.insert(j+2*k);
    LS.push_back(S);
  }
  map<int,int> freq;
  map2freq(LS,freq);
  map<int,int>::iterator t=freq.begin();
  while (t!=freq.end()) {
    cout << "freq[" << t->first << "]: " << t->second << endl;
    t++;
  }
#endif
  
  // ev.eval2(map2freq,vrbs);
  // ev.eval3(ispermfun,vrbs);
  // ev.eval4(filtbyimg,vrbs);
  
  return 0;
}
