#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include "eval.hpp"

using namespace aed;
using namespace std;
// COMPLETAR DNI y NOMBRE AQUI:
// Nombre: Esponja, Bob
int DNI=23456789;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class sumhasher_t : public inthasher_t {
  int state;
public:
  sumhasher_t() : state(0) {}
  void reset() { state=0; }
  void hash(int x) { state += x; }
  int val() { return state; }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class cubhasher_t : public inthasher_t {
  int state;
public:
  cubhasher_t() : state(0) {}
  void reset() { state=0; }
  void hash(int x) {
    int tmp=56789234+12345678*state+x;
    state = tmp*tmp*tmp;
  }
  int val() { return state; }
};

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int hash_set(set<int> &s,inthasher_t &h) {
  h.reset();
  for (auto &x : s) h.hash(x);
  return h.val();
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int count_different(vector<vector<int>> &vectors,
                    inthasher_t &hasher) {
  set<int> hashes;
  int nvec = vectors.size();
  for (int j=0; j<nvec; j++) {
    vector<int> &w = vectors[j];
    hasher.reset();
    for (auto &x : w) hasher.hash(x);
    auto h = hasher.val();
    hashes.insert(h);
  }
  return hashes.size();
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main_hasher() {
  int nvec=100000,m=4,max=100;
  vector<vector<int>> vectors(nvec);
  for (int j=0; j<nvec; j++) {
    auto &w = vectors[j];
    for (int k=0; k<m; k++) w.push_back(rand()%max);
  }
  cubhasher_t chasher;
  printf("nvec %d, different %d\n",
         nvec,count_different(vectors,chasher));
  // aedhasher_t ahasher;
  // sumhasher_t shasher;
  // test_hasher(vectors,ahasher);
  // test_hasher(vectors,shasher);
  return 0;
}


//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool number_graph_pred(graph_t &g,inthasher_t &hasher,
                       vector<int> &vrtxhash) {
  int vrbs=0;
  int nv = g.size();
  vrtxhash.clear();
  vrtxhash.resize(nv,0);
  // Initialize with the hashes of the valences
  for (int j=0; j<nv; j++) {
    hasher.reset();
    hasher.hash(g[j].size());
    vrtxhash[j] = hasher.val();
  }

  int niter=2*nv;
  for (int iter=0; iter<niter; iter++) {
    vector<int> vh2(nv);
    // The new hash of a vertex is the hash of the old hash
    // of the node and the hash of the sorted hashes of the
    // neighbors
    for (int k=0; k<nv; k++) {
      hasher.reset();
      hasher.hash(vrtxhash[k]);
      int nngbr = g[k].size();
      vector<int> hngbr(nngbr);
      set<int> &ngbrs = g[k];
      int indx=0;
      for (auto l : ngbrs) 
        hngbr[indx++] = vrtxhash[l];
      sort(hngbr.begin(),hngbr.end());
      for (int j=0; j<nngbr; j++)
        hasher.hash(hngbr[j]);
      vh2[k] = hasher.val();
      if (vrbs)
        printf("vrtx %d, hash %d\n",k,vh2[k]);
    }
    vrtxhash.swap(vh2);
    vh2 = vrtxhash;
    if (vrbs) {
      sort(vh2.begin(),vh2.end());
      hasher.reset();
      for (int k=0; k<nv; k++) 
        hasher.hash(vh2[k]);
    
      printf("iter %d, hash %d\n",iter,hasher.val());
    }
  }
  set<int> hset;
  for (auto &h : vrtxhash) hset.insert(h);
  return int(hset.size())==nv;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void number_graph(graph_t &g,inthasher_t &hasher,
                  vector<int> &vrtxhash) {
  number_graph_pred(g,hasher,vrtxhash);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void isograph(graph_t &g1,graph_t &g2,
              inthasher_t &hasher,vector<int> &perm) {
  int nv=g1.size();
  vector<int> vh1,vh2;
  number_graph(g1,hasher,vh1);
  number_graph(g2,hasher,vh2);
  map<int,int> h2v;
  for (int j=0; j<nv; j++)
    h2v[vh2[j]]=j;
  perm.resize(nv,0);
  for (int j=0; j<nv; j++)
    perm[j] = h2v[vh1[j]];
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mksumhasher(unique_ptr<inthasher_t> &p) {
  p.reset(new sumhasher_t);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mkcubhasher(unique_ptr<inthasher_t> &p) {
  p.reset(new cubhasher_t);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // sumhasher
  eval1_t eval1;
  eval1.evfun();

  // cubhasher
  eval2_t eval2;
  eval2.evfun();

  // count_different
  eval3_t eval3;
  eval3.evfun();

  // isograph
  eval4_t eval4;
  eval4.evfun();
  
  return 0;
}

