#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include <Eigen/Dense>
#include "eval.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void splitmono(vector<int> &v,vector< vector<int> > &vmono) {
  vmono.clear();
  int j=0, N=v.size();
  while (j<N) {
    vector<int> w;
    w.push_back(v[j++]);
    while (j<N && v[j]>=v[j-1]) w.push_back(v[j++]);
    vmono.push_back(w);
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool sumzero(set<int> &S) {
  int sum=0;
  for (auto q:S) sum += q;
  return sum==0;
}

bool hassize2(set<int> &S) { return S.size()==2; }

bool hasodd(set<int> &S) {
  for (auto q:S) if (q%2) return true;
  return false;
}

bool allpositive(set<int> &S) {
  for (auto q:S) if (q<=0) return false;
  return true;
}

void filtersets(vector< set<int> > &vset,
                vector< set<int> > &vsetout,
                pred_t pred) {
  vsetout.clear();
  int N = vset.size();
  for (int j=0; j<N; j++)
    if (pred(vset[j]))
      vsetout.push_back(vset[j]);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool inside(MatrixXd &xvrtx, VectorXd &xprobe) {
  int ndim = xprobe.size();
  VectorXd rhs(ndim+1),alpha(ndim+1);
  MatrixXd C(ndim+1,ndim+1);
  C.block(0,0,ndim,ndim+1) = xvrtx;
  for (int k=0; k<ndim+1; k++) C(ndim,k)=1.0;
  rhs.segment(0,ndim) = xprobe;
  rhs(ndim) = 1.0;
  alpha = C.lu().solve(rhs);
  for (int j=0; j<ndim+1; j++)
    if (alpha(j)<0.0 || alpha(j)>1) return false;
  return true;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // Para chequear ejercicio 1
  int vrbs=0;
  Eval ev;
  int seed = 123;
  int h1=0,h2=0,h3=0;

  ev.eval<1>(splitmono,vrbs);
  h1 = ev.evalr<1>(splitmono,seed,vrbs);

  ev.eval<2>(filtersets,vrbs);
  h2 = ev.evalr<2>(filtersets,seed,vrbs);

  ev.eval<3>(inside,vrbs);
  h3 = ev.evalr<3>(inside,seed,vrbs);

  printf("S=%03d -> H1=%03d H2=%03d H3=%03d\n",
         seed,h1,h2,h3);
  return 0;
}
