#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include <Eigen/Dense>
#include <vector>
#include "eval.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void mapvec(map<int,int> &M,bool (*pred)(int),vector<int> &v,
             vector<int> &w) {
  w.clear();
  for (auto q:v) {
    int x = M[q];
    if (pred(x)) w.push_back(x);
  }
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void only1(vector< set<int> >&VS,set<int> &DS) {
  DS.clear();
  map<int,int> M;
  for (auto s:VS) 
    for (auto x:s) M[x]++;
  for (auto x:M)
    if (x.second==1)
      DS.insert(x.first);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool reciprocal(map<int,int> &M1,map<int,int> &M2) {
  if (M1.size()!=M2.size()) return false;
  for (auto q:M1) {
    int x = q.first, y=q.second;
    if (M2.find(y)==M2.end()) return false;
    if (M2[y]!=x) return false;
  }
  return true;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
class collabr_t : public player_t {
public:
  move_t move(vector<move_t> &me,vector<move_t> &you) { return COLLAB; }
  string state() {return string("collaborator"); }
};

class cheater_t : public player_t {
public:
  move_t move(vector<move_t> &me,vector<move_t> &you) { return CHEAT; }
  string state() {return string("cheater"); }
};

class copycat_t : public player_t {
public:
  move_t move(vector<move_t> &me,vector<move_t> &you) {
    if (you.size()>0) return you.back();
    else return COLLAB;
  }
  string state() {return string("copycat"); }
};

class grudger_t : public player_t {
public:
  move_t move(vector<move_t> &me,vector<move_t> &you) {
    unsigned int n = you.size();
    for (int j=0; j<n; j++)
      if (you[j] == CHEAT) return CHEAT;
    return COLLAB;
  }
  string state() {return string("grudger"); }
};

class copysmart_t : public player_t {
public:
  move_t move(vector<move_t> &me,vector<move_t> &you) {
    int ncheck=2;
    unsigned int n = you.size();
    if (n<ncheck) return COLLAB;
    for (int j=0; j<ncheck; j++)
      if (you[n-1-j] == COLLAB) return COLLAB;
    return CHEAT;
  }
  string state() {return string("copysmart"); }
};

move_t reverse(move_t mv) {
  return (mv==CHEAT? COLLAB : CHEAT);
}

int DBG = 0;
void game(player_t &p1,player_t &p2,
          int nmoves,
          vector<move_t> &moves1,vector<move_t> &moves2,
          int &score1,int &score2) {
  double noise = 0.0;
  score1=0;
  score2=0;
  moves1.clear();
  moves2.clear();
  for (int j=0; j<nmoves; j++) {
    move_t mv1 = p1.move(moves1,moves2);
    move_t mv2 = p2.move(moves2,moves1);
    if (drand()<noise) mv1 = reverse(mv1);
    if (drand()<noise) mv2 = reverse(mv2);
    if (DBG) printf("mv1 %d, mv2 %d\n",mv1,mv2);
    moves1.push_back(mv1);
    moves2.push_back(mv2);
    if (mv1==COLLAB && mv2==COLLAB)     { score1 +=2; score2 +=2; }
    else if (mv1==COLLAB && mv2==CHEAT) { score1 +=-1; score2 +=3; }
    else if (mv1==CHEAT && mv2==COLLAB) { score1 +=3; score2 +=-1; }
    else if (mv1==CHEAT && mv2==CHEAT)  { score1 +=0; score2 +=0; }
  }
}

#if 0
void trust_game2() {
#define P1 p1.state().c_str()
#define P2 p2.state().c_str()
#if 1
  vector< unique_ptr<player_t> > players;
  players.push_back(unique_ptr<player_t>(new collabr_t));
  players.push_back(unique_ptr<player_t>(new cheater_t));
  players.push_back(unique_ptr<player_t>(new copycat_t));
  players.push_back(unique_ptr<player_t>(new copysmart_t));
  players.push_back(unique_ptr<player_t>(new grudger_t));
  int nplayers = players.size(), nmoves=5000;
  for (int j=0; j<nplayers; j++) {
    player_t &p1 = *players[j];
    int total_score=0, total_moves=0;
    for (int k=0; k<nplayers; k++) {
      player_t &p2 = *players[k];
      int score1,score2;
      game(*players[j],*players[k],nmoves,score1,score2);
      total_score += score1;
      total_moves += nmoves;
      printf("Result: %s: score %d (%.3f per move), "
             "| %s  score: %d (%.3f per move)\n",
             P1,score1,double(score1)/nmoves,
             P2,score2,double(score2)/nmoves);
    }
    printf("===== Total score %s: %d (%.3f/move)\n",P1,total_score,double(total_score)/total_moves);
  }
#else
  player_t
    &p1 = copycat,
    &p2 = grudger;
  int nmoves=5000,score1,score2;
  game(p1,p2,nmoves,score1,score2);
  printf("p1 %s: score %d (%.3f per move), "
         "p2 %s | score: %d (%.3f per move)\n",
         P1,score1,double(score1)/nmoves,
         P2,score2,double(score2)/nmoves);
#endif
}
#endif

void trust_game(map<string,void*> &data) {
  data["collabr"] = new collabr_t;
  data["cheater"] = new cheater_t;
  data["copycat"] = new copycat_t;
  data["copysmart"] = new copysmart_t;
  data["grudger"] = new grudger_t;
  data["game"] = (void*)game;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
bool is_insidecc(MatrixXd &X,VectorXd &xprobe) {
  int ndim = xprobe.size();
  assert(X.rows()==ndim);
  assert(X.cols()==ndim+1);
  MatrixXd A(ndim,ndim);
  VectorXd rhs(ndim);
  for (int j=0; j<ndim; j++) {
    VectorXd dx1j,x1jc;
    dx1j = X.col(j+1)-X.col(0);
    x1jc = 0.5*(X.col(j+1)+X.col(0));
    A.row(j) = dx1j.transpose();
    rhs(j) = dx1j.dot(x1jc);
  }
#define SHV(x) cout << #x ": " << x << endl;
  VectorXd xc = A.lu().solve(rhs);
  // SHV(A);
  // SHV(rhs);
  // SHV(xc);
  VectorXd dx0 = X.col(0)-xc;
  VectorXd dxp = xprobe-xc;
  double
    r2 = dx0.dot(dx0),
    adxp = dxp.dot(dxp);
  return adxp<r2;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void lsqfit(VectorXd &X,VectorXd &Y,basis_t basis,
            int nterms,VectorXd &a) {
  int n = X.rows();
  MatrixXd Phi(n,nterms);
  for (int j=0; j<n; j++)
    for (int k=0; k<nterms; k++) 
      Phi(j,k) = basis(X(j),k);
  MatrixXd H = Phi.transpose()*Phi;
  VectorXd rhs = Phi.transpose()*Y;
  a = H.lu().solve(rhs);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // Para chequear ejercicio 1
  int vrbs=0;
  Eval ev;
  int seed = 123;
  int h1=0,h2=0,h3=0,h4=0,h5=0,h6=0;

#if 0
  // User typical call
  ev.eval<1>(mapvec,vrbs);
  h1 = ev.evalr<1>(mapvec,seed,vrbs);

  ev.eval<2>(only1,vrbs);
  h2 = ev.evalr<2>(only1,seed,vrbs);

  ev.eval<3>(reciprocal,vrbs);
  h3 = ev.evalr<3>(reciprocal,seed,vrbs);

  ev.eval<4>(is_insidecc,vrbs);
  h4 = ev.evalr<4>(is_insidecc,seed,vrbs);

  ev.eval<5>(trust_game,vrbs);
  h5 = ev.evalr<5>(trust_game,seed,vrbs);

  printf("S=%03d -> H1=%03d H2=%03d H3=%03d H4=%03d H5=%03d\n",
         seed,h1,h2,h3,h4,h5);
#elif 1
  // Check a specific task
  ev.eval<6>(lsqfit,vrbs);
  // h6 = ev.evalr<6>(lsqfit,seed,vrbs);
#elif 0
  // Generate seed/hash table
  for (int j=0; j<50; j++) {
    seed = 100+rand()%900;
    if (j==0) seed=123;
    h1 = ev.evalr<1>(mapvec,seed,vrbs);
    h2 = ev.evalr<2>(only1,seed,vrbs);
    h3 = ev.evalr<3>(reciprocal,seed,vrbs);
    h4 = ev.evalr<4>(is_insidecc,seed,vrbs);
    h5 = ev.evalr<5>(trust_game,seed,vrbs);
    printf("S=%03d -> H1=%03d H2=%03d H3=%03d H4=%03d H5=%03d\n",
           seed,h1,h2,h3,h4,h5);
  }
#elif 0
  do {
    ev.eval<1>(mapvec,vrbs);
    h1 = ev.evalr<1>(mapvec,seed,vrbs);

    ev.eval<2>(only1,vrbs);
    h2 = ev.evalr<2>(only1,seed,vrbs);

    ev.eval<3>(reciprocal,vrbs);
    h3 = ev.evalr<3>(reciprocal,seed,vrbs);

    ev.eval<4>(is_insidecc,vrbs);
    h4 = ev.evalr<4>(is_insidecc,seed,vrbs);

    ev.eval<5>(trust_game,vrbs);
    h5 = ev.evalr<5>(trust_game,seed,vrbs);

    printf("S=%03d -> H1=%03d H2=%03d H3=%03d H4=%03d H5=%03d\n",
           seed,h1,h2,h3,h4,h5);
    
    printf("\n\nIngrese un valor para la semilla:");
  } while (cin>>seed);
#endif

  return 0;
}
