#define USECHRONO
#include <cassert>
#include <climits>
#include <cstdlib>
#include <stack>
#include <Eigen/Dense>
#include <vector>
#include "eval.hpp"

using namespace aed;
using namespace std;
using namespace Eigen;

vector<int> findsum(vector<int>& v,int S) {
  int n=v.size();
  vector<int> w;
  for (int j=0; j<n; j++) {
    int s=0;
    for (int k=j; k<n; k++) {
      s+=v[k];
      if (s==S) {
        for (int l=j; l<=k; l++) w.push_back(v[l]);
        return w;
      }
    }
  }
  return w;
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
void xinyect(vector<set<int> > &VS_IN,
             int (*func)(int), vector<set<int> > &VS_OUT){
  for (auto &S:VS_IN) {
    set<int> Aux;
    for (int x : S)
      Aux.insert(func(x));
    if (Aux.size()==S.size())
      VS_OUT.push_back(S);
  }
}

class lin_transform_t {
public:
  virtual int dim()=0;
  virtual void transform(VectorXd &x,VectorXd &y)=0;
};

void getjaco(lin_transform_t &lt,MatrixXd &A,VectorXd &b) {
  int n=lt.dim();
  VectorXd x(n),y(n);
  A.resize(n,n);
  x.fill(0);
  lt.transform(x,b);
  
  for (int j=0; j<n; j++) {
    x.fill(0);
    x(j) = 1;
    lt.transform(x,y);
    A.col(j) = y-b;
  }
}

class shift_t : public lin_transform_t {
public:
  shift_t(int n_,double c_) : n(n_), c(c_) { }
  int n;
  double c;
  int dim() { return n; }
  void transform(VectorXd &x,VectorXd &y) {
#if 0    
    VectorXd cv(x);
    cv.fill(c);
    y = x + cv;
#else
    y = x;
    y.array() += c;
#endif
  }
};

class ident_t : public lin_transform_t {
public:
  ident_t(int n_) : n(n_) { }
  int n;
  int dim() { return n; }
  void transform(VectorXd &x,VectorXd &y) {
    y = x;
  }
};

class scale_t : public lin_transform_t {
public:
  scale_t(int n_,double c_) : n(n_), c(c_) { }
  int n;
  double c;
  int dim() { return n; }
  void transform(VectorXd &x,VectorXd &y) {
    y = c*x;
  }
};

class rotate_t : public lin_transform_t {
public:
  rotate_t(int n_) : n(n_) { }
  int n;
  int dim() { return n; }
  void transform(VectorXd &x,VectorXd &y) {
    y = x;
    for (int j=0; j<n; j++) {
      int k = (j+1)%n;
      y(k) = x(j);
    }
  }
};

void check1(lin_transform_t &lt) {
  cout << "================\n";
  MatrixXd A;
  VectorXd b;
  getjaco(lt,A,b);
  cout << "A: " << endl;
  cout << A << endl;
  cout << "b: " << endl;
  cout << b << endl;
}

void check_getjaco() {
  shift_t shift(10,3.0); check1(shift);
  ident_t ident(5); check1(ident);
  scale_t scale(4,8.3); check1(scale);
  rotate_t rotate(5); check1(rotate);
}

//---:---<*>---:---<*>---:---<*>---:---<*>---:---<*>
int main() {
  // Para chequear ejercicio 1
  int vrbs=0;
  Eval ev;
  int seed = 123;

  // ev.eval<1>(findsum,1);
  // ev.eval<2>(xinyect,vrbs);
  check_getjaco();

  return 0;
}
